# Distributed under the terms of GPL version 2 or any later
# Copyright (C) Kristopher Tate / BlueBridge Technologies Group 2005
# Copyright (C) Alexey Nezhdanov 2004

# Stream-level features for xmppd.py

from xmpp import *
from xmppd import *
import socket,thread,os,base64
from tlslite.api import *

class TLS(PlugIn):
    """ 3.                        <features/>
        4. <starttls/>
        5.                        <proceed/> / <failure/>
        -- NEW STREAM / connection close --
        """
    NS=NS_TLS
    def plugin(self,server):
        server.Dispatcher.RegisterHandler('starttls',self.starttlsHandler,xmlns=NS_TLS)
        server.Dispatcher.RegisterHandler('proceed',self.proceedfailureHandler,xmlns=NS_TLS)
        server.Dispatcher.RegisterHandler('failure',self.proceedfailureHandler,xmlns=NS_TLS)
        server.Dispatcher.RegisterHandler('features',self.FeaturesHandler,xmlns=NS_STREAMS)

    def starttlsHandler(self,session,stanza):
        if 'tls' in session.features:
            session.send(Node('failure',{'xmlns':NS_TLS}))
            self.DEBUG('TLS startup failure: already started.','error')
            session.unfeature(NS_TLS)
            raise NodeProcessed
        if self._owner.__dict__.has_key('sslcertfile'): certfile=self._owner.sslcertfile
        else: certfile=None
        if self._owner.__dict__.has_key('sslkeyfile'): keyfile=self._owner.sslkeyfile
        else: keyfile=certfile
        try: open(certfile) ; open(keyfile)
        except: certfile=None
        if not certfile or not keyfile:
            session.send(Node('failure',{'xmlns':NS_TLS}))
            self.DEBUG('TLS startup failure: can\'t find SSL cert/key file[s].','error')
            session.unfeature(NS_TLS)    # do not declare TLS anymore
            session.stop_feature(NS_TLS) # TLS finished, let another features start
        else:
            session.send(Node('proceed',{'xmlns':NS_TLS}))
            self.startservertls(session)
        raise NodeProcessed

    def startservertls(self,session):
        session._owner.unregistersession(session)
        self._startservertls(session)

    def _startservertls(self,session):
        try:
            cert=open(self._owner.sslcertfile).read()
            key=open(self._owner.sslkeyfile).read()
        except:
            session.unfeature(NS_TLS)
            session.terminate_stream(STREAM_INTERNAL_SERVER_ERROR)
            return

        x509 = X509()
        x509.parse(cert)
        certChain = X509CertChain([x509])
        privateKey = parsePEMKey(key, private=True)
        connection = TLSConnection(session._sock)
        try: connection.handshakeServer(certChain=certChain, privateKey=privateKey, reqCert=False)
        except:
            session.terminate_stream(unregister=0)
            return

        session._sslObj = connection
        session._recv = connection.read
        session._send = connection.send

        session.feature(NS_TLS)      # TLS enabled, do not declare it anymore
        session.stop_feature(NS_TLS) # TLS finished, let another features start
        session.StartStream()
        session._owner.registersession(session)

    def proceedfailureHandler(self,session,stanza):
        if stanza.getName()<>'proceed':
            self.DEBUG('TLS can not be started. Giving up.','error')
            session.unfeature(NS_TLS)
            raise NodeProcessed
        self.DEBUG('Starting client-mode TLS.','ok')
        try: session._sslObj = socket.ssl(session._sock, None, None)
        except:
            session.set_socket_state(SOCKET_DEAD)
            self.DEBUG('TLS failed. Terminating session','error')
            session.terminate_stream()
            raise NodeProcessed
        session._recv = session._sslObj.read
        session._send = session._sslObj.write

        session.feature(NS_TLS)      # TLS enabled, do not declare it anymore
        session.stop_feature(NS_TLS) # TLS finished, let another features start
        session.StartStream()
        raise NodeProcessed

    def FeaturesHandler(self,session,stanza):
        if NS_TLS in session.features: return     # already started. do nothing
        if session.feature_in_process: return     # some other feature is already underway
        if not stanza.getTag('starttls',namespace=NS_TLS):
            self.DEBUG("TLS unsupported by remote server; Doing nothing -- Prob. jabber.org?",'warn')
            return
        else:
            self.DEBUG("TLS supported by remote server. Requesting TLS start.",'ok')
            session.start_feature(NS_TLS)
            session.send(Node('starttls',{'xmlns':NS_TLS}))
        raise NodeProcessed

class SASL(PlugIn):
    NS=NS_SASL
    """ 3.                        <features/>
        4. <auth/>
        5.                        <challenge/> / <failure/>
        6. <response/>
        7.                        <challenge/> / <failure/>
        8. <response/> / <abort/>
        9.                        <success/>   / <failure/>
        feature SASL, unfeature TLS
        -- NEW STREAM on success --

        What to do on failure (remote server rejected us)?
        Probably drop the stream, mark this server as unreachable for several hours and notify admin?
        If client supplied wrong credentials allow him to retry (configurable number of times).
        """

    def plugin(self,server):
        server.Dispatcher.RegisterNamespaceHandler(NS_SASL,self.SASLHandler)
#        server.Dispatcher.RegisterHandler('features',self.FeaturesHandler,xmlns=NS_STREAMS)
        self.auths=auth.getSASLAuthMechanisms(self.DEBUG)
        self.mechanisms=self.auths.keys()
        self.authmech = None

    def commit_auth(self,session,authzid,payload=[]):
        session.send(Node('success',{'xmlns':NS_SASL}, payload))
        session.feature(NS_SASL)
        session.unfeature(NS_TLS)
        session.sasl['next']=[]
        session.StartStream()
        session.peer=authzid.lower()
        if session.xmlns==NS_CLIENT: session.set_session_state(SESSION_AUTHED)
        else: session.set_session_state(SESSION_BOUND)
        self.DEBUG('Peer %s successfully authenticated'%authzid,'ok')

    def reject_auth(self,session,authzid='unknown',reason='not-authorized'):
        session.send(Node('failure',{'xmlns':NS_SASL},[Node(reason)]))
        session.sasl['retries']=session.sasl['retries']-1
        if session.sasl['retries']<=0: session.terminate_stream()
        self.DEBUG('Peer %s failed to authenticate because %s'%(authzid,reason),'error')

    def SASLHandler(self,session,stanza):
        """simple username: servername _or_ node@servername : 6.1 (6)
        """
        if NS_SASL in session.features:
            self.DEBUG('Already authorized. Ignoring SASL stanza.','error')
            raise NodeProcessed
        if not session.__dict__.has_key('sasl'):
            session.sasl={'retries':3}
        if not session.sasl.has_key('next'):
            session.sasl={'retries':session.sasl['retries']}
            if session.TYP=='server': session.sasl['next']=['auth','abort']
            else: session.sasl['next']=['challenge','success','failure']
        if stanza.getName() not in session.sasl['next']:
            # screwed SASL implementation on the other side. terminating stream
            session.terminate_stream(ERR_BAD_REQUEST)
            raise NodeProcessed
        #=================== preparation ===============================================
        data=stanza.getData()
        if data:
            try: data=base64.decodestring(data)
            except:
                session.terminate_stream(ERR_BAD_REQUEST)
                raise NodeProcessed
        else:
            data=None
        self.DEBUG('Got challenge: '+`data`,'ok')
        #=================== SASL server begin =========================================
        if stanza.getName()=='auth':
            session.sasl['next']=['response','abort']
            # client requested some mechanism. May be ever provided credentials already.
            mec=stanza['mechanism']
            session.sasl['mechanism']=mec
            self.authmech = None
            if self.auths.has_key(mec):
                self.authmech = self.auths[mec]()
        if stanza.getName()=='auth' or stanza.getName()=='response':
            if self.authmech:
                result, authzid, extra = self.authmech.AuthHandler(session, self, data)
                if result == 'success':
                    if extra != None:
                        if extra:
                            payload = base64.encodestring(extra).replace('\n','')
                        else:
                            payload = '='
                    else:                            
                        payload=[]
                    self.commit_auth(session,authzid,payload)
                elif result == 'fail':
                    session.sasl['next']=['auth','abort']
                    self.reject_auth(session,authzid,extra)
                elif result == 'challenge':
                    session.sasl['next']=['response','abort']
                    if extra != None:
                        if extra:
                            payload = base64.encodestring(extra).replace('\n','')
                        else:
                            payload = '='
                    else:
                        payload=[]
                    session.send(Node(result,{'xmlns':NS_SASL},payload))
                else:
                    self.DEBUG('Bad AuthHandler result: %s'%result,'error')
                    session.terminate_stream(ERR_BAD_REQUEST)
            else:
                session.terminate_stream(Node('failure',{'xmlns':NS_SASL},[Node('invalid-mechanism')]))
            raise NodeProcessed
        elif stanza.getName()=='abort':
            session.sasl['next']=['auth']
            self.authmech = None
            self.reject_auth(session,reason='aborted')
        #=================== SASL client begin =========================================
        elif stanza.getName()=='challenge':
            session.sasl['next']=['challenge','success','failure']
        elif stanza.getName()=='success':
            session.sasl['next']=[]
            session.startsasl='success'
            self.DEBUG('Successfully authenticated with remote server.','ok')
            session.StartStream()
        elif stanza.getName()=='failure':
            session.sasl['next']=['challenge','success','failure']
            session.startsasl='failure'
            try: reason=challenge.getChildren()[0]
            except: reason=challenge
            self.DEBUG('Failed SASL authentification: %s'%reason,'error')
        else:
            session.terminate_stream(ERR_BAD_REQUEST)
        raise NodeProcessed

class Bind(PlugIn):
    NS=NS_BIND
    def plugin(self,server):
        server.Dispatcher.RegisterHandler('iq',self.bindHandler,typ='set',ns=NS_BIND,xmlns=NS_CLIENT)

    def bindHandler(self,session,stanza):
        if session.xmlns<>NS_CLIENT or session.__dict__.has_key('resource'):
            session.send(Error(stanza,ERR_SERVICE_UNAVAILABLE))
        else:
            if session._session_state<SESSION_AUTHED:
                session.terminate_stream(STREAM_NOT_AUTHORIZED)
                raise NodeProcessed
            resource=stanza.getTag('bind',namespace=NS_BIND).T.resource.getData()
            if not resource: resource=session.ID
            fulljid="%s/%s"%(session.peer,resource)
            session.peer=fulljid
            s=self._owner.deactivatesession(fulljid)
            if s: s.terminate_stream(STREAM_CONFLICT)
            rep=stanza.buildReply('result')
            rep.T.bind.setNamespace(NS_BIND)
            rep.T.bind.T.jid=fulljid
            session.send(rep)
            session.set_session_state(SESSION_BOUND)
        raise NodeProcessed

class Session(PlugIn):
    NS=NS_SESSION
    def plugin(self,server):
        server.Dispatcher.RegisterHandler('iq',self.sessionHandler,typ='set',ns=NS_SESSION,xmlns=NS_CLIENT)

    def sessionHandler(self,session,stanza):
        if session._session_state<SESSION_AUTHED:
            session.terminate_stream(STREAM_NOT_AUTHORIZED)
            raise NodeProcessed
        if session.xmlns<>NS_CLIENT \
          or session._session_state<SESSION_BOUND \
          or self._owner.getsession(session.peer)==session:
            session.send(Error(stanza,ERR_SERVICE_UNAVAILABLE))
        else:
            session.set_session_state(SESSION_OPENED)
            session.send(stanza.buildReply('result'))
        raise NodeProcessed
